/*
 * Copyright 1994 by OpenVision Technologies, Inc.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of OpenVision not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission. OpenVision makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 * 
 * OPENVISION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL OPENVISION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef macintosh
#define _MACINTOSH /* GSS headers expect this */
#endif

#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifndef macintosh
#	include <sys/types.h>
#else
#	include <mit-sock.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#include <errno.h>
#ifndef macintosh
#	include <sys/stat.h>
#else
#	include <stat.h>
#endif
#include <fcntl.h>

//#include <gssapi/gssapi_generic.h>
#include "gss-misc.h"

#include "gss-client.h"
#include "krb5.h"

/* From the GSSAPI header! */
#define KG_EMPTY_CCACHE                          (39756044L)

#ifndef macintosh
#define socket_close(s)	close(s)
#endif

static const gss_OID_desc oids[] = {
   {10, "\052\206\110\206\367\022\001\002\001\004"},
};

static const gss_OID_desc * nt_service_name = oids+0;

/*
 * Function: connect_to_server
 *
 * Purpose: Opens a TCP connection to the name host and port.
 *
 * Arguments:
 *
 * 	host		(r) the target host name
 * 	port		(r) the target port, in host byte order
 *
 * Returns: the established socket file desciptor, or -1 on failure
 *
 * Effects:
 *
 * The host name is resolved with gethostbyname(), and the socket is
 * opened and connected.  If an error occurs, an error message is
 * displayed and -1 is returned.
 */
static int connect_to_server(host, port, ret_s, errstring)
	char *host;
	unsigned short port;
	int *ret_s;
	char *errstring;
{
     struct sockaddr_in saddr;
     struct hostent *hp;
     int s;
#ifdef _WIN32
     size_t ws_err;
#endif
     
     if ((hp = gethostbyname(host)) == NULL) {
     	  strcpy(errstring, "Unknown host: ");
	  strncat(errstring, host, 50);
	  return -1;
     }
     
     saddr.sin_family = hp->h_addrtype;
     memcpy((char *)&saddr.sin_addr, hp->h_addr, sizeof(saddr.sin_addr));
     saddr.sin_port = htons(port);

     if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#ifdef _WIN32
        ws_err = WSAGetLastError();
	errno = ws_err;
#endif
	sprintf(errstring, "Couldn't create socket (%d)", errno);
	return errno;
     }

     if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
#ifdef _WIN32
        ws_err = WSAGetLastError();
	errno = ws_err;
#endif
	strcpy(errstring, "Couldn't connect to Krb5 connect server on host ");
	strncat(errstring, host, 50);
	return errno;
     }

     *ret_s = s;
     return 0;
}

/*
 * Function: client_establish_context
 *
 * Purpose: establishes a GSS-API context with a specified service and
 * returns the context handle
 *
 * Arguments:
 *
 * 	s		(r) an established TCP connection to the service
 * 	service_name	(r) the ASCII service name of the service
 * 	context		(w) the established GSS-API context
 *	ret_flags	(w) the returned flags from init_sec_context
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 * 
 * service_name is imported as a GSS-API name and a GSS-API context is
 * established with the corresponding service; the service should be
 * listening on the TCP connection s.  The default GSS-API mechanism
 * is used, and mutual authentication and replay detection are
 * requested.
 * 
 * If successful, the context handle is returned in context.  If
 * unsuccessful, the GSS-API error messages are displayed on stderr
 * and -1 is returned.
 */
static int client_establish_context(s, service_name, deleg_flag, oid,
				    gss_context, estr, ret_flags)
     int s;
     char *service_name;
     gss_OID oid;
     OM_uint32 deleg_flag;
     gss_ctx_id_t *gss_context;
     char *estr;
     OM_uint32 *ret_flags;
{
     gss_buffer_desc send_tok, recv_tok, *token_ptr, msg;
     gss_name_t target_name;
     OM_uint32 maj_stat, min_stat, maj_1, min_1;
     int retval, len;
     unsigned int msg_ctx=0;

     /*
      * Import the name into target_name.  Use send_tok to save
      * local variable space.
      */
     send_tok.value = service_name;
     send_tok.length = strlen(service_name) + 1;
     maj_stat = gss_import_name(&min_stat, &send_tok,
				(gss_OID) nt_service_name, &target_name);
     if (maj_stat != GSS_S_COMPLETE) {
	  display_status(0, "parsing name", maj_stat, min_stat);
	  return -1;
     }
     
     /*
      * Perform the context-establishement loop.
      *
      * On each pass through the loop, token_ptr points to the token
      * to send to the server (or GSS_C_NO_BUFFER on the first pass).
      * Every generated token is stored in send_tok which is then
      * transmitted to the server; every received token is stored in
      * recv_tok, which token_ptr is then set to, to be processed by
      * the next call to gss_init_sec_context.
      * 
      * GSS-API guarantees that send_tok's length will be non-zero
      * if and only if the server is expecting another token from us,
      * and that gss_init_sec_context returns GSS_S_CONTINUE_NEEDED if
      * and only if the server has another token to send us.
      */
     
     token_ptr = GSS_C_NO_BUFFER;
     *gss_context = GSS_C_NO_CONTEXT;

     do {
	  maj_stat =
	       gss_init_sec_context(&min_stat,
				    GSS_C_NO_CREDENTIAL,
				    gss_context,
				    target_name,
				    oid,
				    GSS_C_MUTUAL_FLAG | GSS_C_REPLAY_FLAG |
							deleg_flag,
				    0,
				    NULL,	/* no channel bindings */
				    token_ptr,
				    NULL,	/* ignore mech type */
				    &send_tok,
				    ret_flags,
				    NULL);	/* ignore time_rec */

	  if (token_ptr != GSS_C_NO_BUFFER) {
	       free(recv_tok.value);
	       recv_tok.value = 0;
	  }
	  if (maj_stat!=GSS_S_COMPLETE && maj_stat!=GSS_S_CONTINUE_NEEDED) {
	      maj_1 = GSS_ERROR(maj_stat);
	      if ((maj_1 == GSS_S_NO_CRED) ||
		  (maj_1 == GSS_S_CREDENTIALS_EXPIRED) ||
		  (maj_1 == GSS_S_DEFECTIVE_CREDENTIAL) ||
		  ((maj_1 == GSS_S_FAILURE) &&
		   ((min_stat == KRB5_FCC_NOFILE) ||
		    (min_stat == KRB5_CC_NOTFOUND) ||
		    (min_stat == KG_EMPTY_CCACHE) ||
		    (min_stat == KRB5KRB_AP_ERR_TKT_EXPIRED))))
		  strcpy(estr, "You need to get fresh Kerberos 5 tickets by using the Kerberos V5 login program");
	      else {
		  maj_1 = gss_display_status(&min_1, min_stat, GSS_C_MECH_CODE,
					     oid, &msg_ctx, &msg);
		  if (!GSS_ERROR(maj_1)) {
		      len = msg.length;
		      if (len > 80)
			  len=80;
		      strcpy(estr, "Kerberos error: ");
		      strncat(estr, msg.value, len);
		      estr[len + sizeof("Kerberos error: ")] = 0;
		      (void) gss_release_buffer(&min_1, &msg);
		  } else
		      display_status(0, "initializing context", maj_stat,
				     min_stat);
	      }
	      

	      (void) gss_release_name(&min_stat, &target_name);
	      return -1;
	  }

	  if (send_tok.length != 0) {
	  	retval = send_token(s, &send_tok);
		if (retval) {
		    (void) gss_release_buffer(&min_stat, &send_tok);
		    (void) gss_release_name(&min_stat, &target_name);
		    return -1;
	       }
	  }
	  (void) gss_release_buffer(&min_stat, &send_tok);
	  
	  if (maj_stat == GSS_S_CONTINUE_NEEDED) {
	       retval = recv_token(s, &recv_tok);
	       if (retval) {
		    (void) gss_release_name(&min_stat, &target_name);
		    return -1;
	       }
	       token_ptr = &recv_tok;
	  }
     } while (maj_stat == GSS_S_CONTINUE_NEEDED);

     (void) gss_release_name(&min_stat, &target_name);
     return 0;
}

/*
 * Function: call_kludge_server
 *
 * Purpose: Call the "kludge" service.
 *
 * Arguments:
 *
 * 	host		(r) the host providing the service
 * 	port		(r) the port to connect to on host
 * 	service_name	(r) the GSS-API service name to authenticate to	
 * 	msg		(r) the message to have "signed"
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 * 
 * call_server opens a TCP connection to <host:port> and establishes a
 * GSS-API context with service_name over the connection.  It then
 * seals msg in a GSS-API token with gss_seal, sends it to the server,
 * reads back a GSS-API signature block for msg from the server, and
 * verifies it with gss_verify.  -1 is returned if any step fails,
 * otherwise 0 is returned.
 */
int call_kludge_server(char *host, unsigned int port, const char *service_name,
		       const char *sid, char **username, char **password,
		       char **errstring)
{
	gss_ctx_id_t context;
	gss_buffer_desc in_buf, out_buf;
	int s, state, retval, conf_state, len;
	OM_uint32 ret_flags;
	OM_uint32 maj_stat, min_stat;
	char		msg[128];
	char		*estr = malloc(128);

	if (!estr)
		return -1;
	*errstring = estr;
	*username = 0;
	*password = 0;

	/* Open connection */
	strcpy(estr, "Connection failed");
	retval = connect_to_server(host, port, &s, estr);
	if (retval)
		return retval;

	/* Establish context */
	strcpy(estr, "Kerberos connection failed.\n");
	if (client_establish_context(s, service_name, 0,
				     GSS_C_NULL_OID, &context, estr,
				     &ret_flags) < 0) {
		(void) socket_close(s);
		return -1;
	}
	strcpy(estr, "kludge server failed");

	in_buf.value = msg;
	msg[0] = 1;		/* Major version number */
	if (sid && sid[0]) {
		msg[1] = 1;	/* Minor version number */
		in_buf.length = 2;
		len = strlen(sid);
		if (len > 16) {
			msg[16] = 0;
			len = 16;
		}
		msg[2] = len;
		strcpy(msg+3, sid);
		in_buf.length = 3 + len;
	} else {
		msg[1] = 0;
		in_buf.length = 2;
	}

	maj_stat = gss_wrap(&min_stat, context, 1, GSS_C_QOP_DEFAULT,
			    &in_buf, &state, &out_buf);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(0, "sealing message", maj_stat, min_stat);
		(void) socket_close(s);
		(void) gss_delete_sec_context(&min_stat, &context,
					      GSS_C_NO_BUFFER);
		strcpy(estr, "Preparing request for password server failed.\n");
      		return -1;
	}

	/* Send to server */
	if (send_token(s, &out_buf) < 0) {
		(void) socket_close(s);
		(void) gss_delete_sec_context(&min_stat, &context,
					      GSS_C_NO_BUFFER);
		strcpy(estr, "Send request to password server failed.\n");
		return -1;
	}
	(void) gss_release_buffer(&min_stat, &out_buf);

	retval = recv_sealed_token(s, context, &out_buf,
				   &conf_state, &maj_stat, &min_stat);
	if (retval) {
		sprintf(estr, "Couldn't get token: %s\n", strerror(retval));
		return(-1);
	}
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(0, "unsealing initial message", maj_stat,
			       min_stat);
		sprintf(estr, "Couldn't decode reply message.\n");
		return(-1);
	}
	if (!conf_state) {
#ifdef unix
		fprintf(stderr, "Warning!  Message not encrypted.\n");
#elif defined (macintosh)
	/* ignore it */
#else
		OkMsgBox("Warning!  Message not encrypted.\n");
#endif
	}

	retval = parse_kludge_token(username, password, &out_buf);
	
	(void) gss_release_buffer(&min_stat, &out_buf);
	(void) socket_close(s);

	if (!retval) {
		free(estr);
		*errstring = 0;
	}
	if (!*username || !**username) {
		if (*password && **password) {
			*errstring = *password;
			*password = 0;
		}
	}
	return retval;
}
