/*
 * gss-misc.c --- Miscellaneous routines
 *
 * Copyright 1997 by Massachusetts Institute of Technology
 * 
 * Copyright 1994 by OpenVision Technologies, Inc.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of OpenVision not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission. OpenVision makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 * 
 * OPENVISION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL OPENVISION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef macintosh
#define _MACINTOSH /* gss headers expect this */
#endif

#include <stdio.h>
#ifndef macintosh
#	include <sys/types.h>
#else
#	include <mit-sock.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include <stdlib.h>
#include <string.h>

//#include <gssapi/gssapi_generic.h> fixme
#include "gss-misc.h"

FILE *display_file;

static void display_status_1
	PROTOTYPE( (FILE *log, char *m, OM_uint32 code, int type) );

#if defined (unix) || defined (macintosh)
/* we have a compatible sockets library on the mac */
/* except that we must use socket_read and socket_write */
#ifdef macintosh
#define write socket_write
#define read socket_read
#endif

static int write_all(int fildes, char *buf, unsigned int nbyte)
{
     int ret;
     char *ptr;

     for (ptr = buf; nbyte; ptr += ret, nbyte -= ret) {
	  ret = write(fildes, ptr, nbyte);
	  if (ret < 0) {
	       if (errno == EINTR)
		    continue;
	       return(ret);
	  } else if (ret == 0) {
	       return(ptr-buf);
	  }
     }

     return(ptr-buf);
}

static int read_all(int fildes, char *buf, unsigned int nbyte)
{
     int ret;
     char *ptr;

     for (ptr = buf; nbyte; ptr += ret, nbyte -= ret) {
	  ret = read(fildes, ptr, nbyte);
	  if (ret < 0) {
	       if (errno == EINTR)
		    continue;
	       return(ret);
	  } else if (ret == 0) {
	       return(ptr-buf);
	  }
     }

     return(ptr-buf);
}

/*
 * Function: send_token
 *
 * Purpose: Writes a token to a file descriptor.
 *
 * Arguments:
 *
 * 	s		(r) an open file descriptor
 * 	tok		(r) the token to write
 *
 * Returns: 0 on success, error number on failure
 *
 * Effects:
 *
 * send_token writes the token length (as a network long) and then the
 * token data to the file descriptor s.  It returns 0 on success, and
 * an error number if an error occurs.
 */
int send_token(s, tok)
     int s;
     gss_buffer_t tok;
{
     int len, ret;

     len = htonl(tok->length);
     ret = write_all(s, (char *) &len, 4);
     if (ret < 0)
	     return errno;
     if (ret != 4)
	     return EIO;
     ret = write_all(s, tok->value, tok->length);
     if (ret < 0)
	     return errno;
     if ((unsigned int) ret != tok->length)
	     return EIO;
     return 0;
}

/*
 * Function: recv_token
 *
 * Purpose: Reads a token from a file descriptor.
 *
 * Arguments:
 *
 * 	s		(r) an open file descriptor
 * 	tok		(w) the read token
 *
 * Returns: 0 on success, error number on failure
 *
 * Effects:
 * 
 * recv_token reads the token length (as a network long), allocates
 * memory to hold the data, and then reads the token data from the
 * file descriptor s.  It blocks to read the length and data, if
 * necessary.  On a successful return, the token should be freed with
 * gss_release_buffer.  It returns 0 on success, or an error number if
 * an error occurs or if it could not read all the data.
 */
int recv_token(s, tok)
	int s;
	gss_buffer_t tok;
{
	int ret;

	ret = read_all(s, (char *) &tok->length, 4);
	if (ret < 0)
		return errno;
	if (ret != 4)
		return EIO;
	tok->length = ntohl(tok->length);
	tok->value = (char *) malloc(tok->length);
	if (tok->value == NULL)
		return ENOMEM;
	ret = read_all(s, (char *) tok->value, tok->length);
	if (ret < 0) {
		free(tok->value);
		return errno;

	}
	if ((unsigned int) ret != tok->length) {
		free(tok->value);
		return EIO;
	}
	return 0;
}
#else
/*+*************************************************************************
** 
** OkMsgBox
** 
** A MessageBox version of printf
** 
***************************************************************************/
void
OkMsgBox (char *format, ...)
{
    char buf[256];								// Message goes into here
    char *args;                                 // Args for printf

    args = (char *) &format + sizeof(format);
    vsprintf (buf, format, args);
    MessageBox(NULL, buf, "", MB_OK);
}

/*+*************************************************************************
** 
** My_perror
** 
** A windows conversion of perror displaying the output into a MessageBox.
** 
***************************************************************************/
void
my_perror (char *msg) {
    char *err;

    err = strerror (errno);

    if (msg && *msg != '\0') 
        OkMsgBox ("%s: %s", msg, err);
    else
        MessageBox (NULL, err, "", MB_OK);
}

/*+
 * Function: send_token
 *
 * Purpose: Writes a token to a file descriptor.
 *
 * Arguments:
 *
 *	s		(r) an open file descriptor
 *	tok		(r) the token to write
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 *
 * send_token writes the token length (as a network long) and then the
 * token data to the file descriptor s.	 It returns 0 on success, and
 * -1 if an error occurs or if it could not write all the data.
 */
int send_token(int s, gss_buffer_t tok) {
    long len;
    size_t ret;
    size_t ws_err;

    len = htonl(tok->length);

    ret = send (s, (char *) &len, 4, 0);        // Send length over the socket
    if (ret < 0) {
		ws_err = WSAGetLastError();
		errno = ws_err;
		my_perror("sending token length");
		OkMsgBox ("Winsock error  %d \n", ws_err);
		return -1;
    } else if (ret != 4) {
	    ws_err = WSAGetLastError();
	    OkMsgBox("sending token length: %d of %d bytes written\nWinsock error = %d\n",
		     ret, 4, ws_err);
	    return -1;
    }

    ret = send (s, tok->value, tok->length, 0); // Send the data
    if (ret < 0) {
	    ws_err = WSAGetLastError();
	    errno = ws_err;
	    my_perror("sending token data");
	    OkMsgBox ("Winsock error  %d \n", ws_err);
	    return -1;
    } else if (ret != tok->length) {
	    ws_err = WSAGetLastError();
	    OkMsgBox ("sending token data: %d of %d bytes written\nWinsock error = %d\n",
		      ret, tok->length, ws_err);
	    return -1;
    }

    return 0;
}

/*+
 * Function: recv_token
 *
 * Purpose: Reads a token from a file descriptor.
 *
 * Arguments:
 *
 *	s		(r) an open file descriptor
 *	tok		(w) the read token
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 * 
 * recv_token reads the token length (as a network long), allocates
 * memory to hold the data, and then reads the token data from the
 * file descriptor s.  It blocks to read the length and data, if
 * necessary.  On a successful return, the token should be freed with
 * gss_release_buffer.	It returns 0 on success, and -1 if an error
 * occurs or if it could not read all the data.
 */
int
recv_token (int s, gss_buffer_t tok) {
    int ret;
    unsigned long len;
	size_t ws_err;

    ret = recv (s, (char *) &len, 4, 0);
    if (ret < 0) {
	    ws_err = WSAGetLastError();
	    errno = ws_err;
	    my_perror("reading token length");
	    OkMsgBox ("Winsock error  %d \n", ws_err);
        return -1;
    } else if (ret != 4) {
	    ws_err = WSAGetLastError();
	    OkMsgBox ("reading token length: %d of %d bytes written\nWinsock error = %d\n",
		      ret, 4, ws_err);
	    return -1;
    }
	  
    len = ntohl(len);
    tok->length = (size_t) len;
    tok->value = (char *) malloc(tok->length);
    if (tok->value == NULL) {
        OkMsgBox ("Out of memory allocating token data\n");
        return -1;
    }

    ret = recv (s, (char *) tok->value, tok->length, 0);
    if (ret < 0) {
	    ws_err = WSAGetLastError();
	    errno = ws_err;
	    my_perror("reading token data");
	    OkMsgBox ("Winsock error  %d \n", ws_err);
	    free(tok->value);
	    return -1;
    } else if ((size_t) ret != tok->length) {
	    ws_err = WSAGetLastError();
	    OkMsgBox ("reading token data: %d of %d bytes written\nWinsock error = %d\n",
		      ret, tok->length, ws_err);
	    free(tok->value);
	    return -1;
    }

    return 0;

}
#endif

int recv_sealed_token(s, context, tok, conf_state, maj_stat, min_stat)
	int s;
	gss_ctx_id_t context;
	gss_buffer_t tok;
	int		*conf_state;
	OM_uint32 	*maj_stat, *min_stat;
{
	gss_buffer_desc	xmit_buf;
	int		retval;

	/* Receive the sealed message token */
	retval = recv_token(s, &xmit_buf);
	if (retval)
		return retval;
	  
	*maj_stat = gss_unwrap(min_stat, context, &xmit_buf, tok,
			       conf_state, (gss_qop_t *) NULL);
	if (*maj_stat != GSS_S_COMPLETE)
		return -1;

	free(xmit_buf.value);
	return 0;
}

int send_sealed_token(s, context, tok, conf_state, maj_stat, min_stat)
	int s;
	gss_ctx_id_t context;
	gss_buffer_t tok;
	int		*conf_state;
	OM_uint32 	*maj_stat, *min_stat;
{
	gss_buffer_desc	out_buf;
	OM_uint32	tmp;
	int		retval, conf_required;
	
	conf_required = *conf_state;
	*maj_stat = gss_wrap(min_stat, context, 1, GSS_C_QOP_DEFAULT,
			    tok, conf_state, &out_buf);
	if (*maj_stat != GSS_S_COMPLETE)
		return -1;

	if (conf_required && !*conf_state) {
		(void) gss_release_buffer(&tmp, &out_buf);
		return -1;
	}

	retval = send_token(s, &out_buf);
	if (retval)
		return retval;

	(void) gss_release_buffer(&tmp, &out_buf);
	return 0;
}

static void display_status_1(log, m, code, type)
     FILE *log;
     char *m;
     OM_uint32 code;
     int type;
{
     OM_uint32 maj_stat, min_stat;
     gss_buffer_desc msg;
     OM_uint32 msg_ctx;
     
     msg_ctx = 0;
     while (1) {
	  maj_stat = gss_display_status(&min_stat, code,
				       type, GSS_C_NULL_OID,
				       &msg_ctx, &msg);
#ifdef unix
	  if (!log)
		  log = stdin;
	  if (log)
		  fprintf(log, "GSS-API error %s: %s\n", m,
			  (char *)msg.value);
#elif defined (macintosh)
		printf ("GSS-API error %s: %s\n", m,
			  (char *)msg.value);
#else
	  OkMsgBox ("GSS-API error %s: %s\n", m,
		    (char *)msg.value);
#endif
	  (void) gss_release_buffer(&min_stat, &msg);
	  
	  if (!msg_ctx)
	       break;
     }
}

/*
 * Function: display_status
 *
 * Purpose: displays GSS-API messages
 *
 * Arguments:
 *
 * 	msg		a string to be displayed with the message
 * 	maj_stat	the GSS-API major status code
 * 	min_stat	the GSS-API minor status code
 *
 * Effects:
 *
 * The GSS-API messages associated with maj_stat and min_stat are
 * displayed on stderr, each preceeded by "GSS-API error <msg>: " and
 * followed by a newline.
 */
void display_status(log, msg, maj_stat, min_stat)
     FILE *log;
     char *msg;
     OM_uint32 maj_stat;
     OM_uint32 min_stat;
{
     display_status_1(log, msg, maj_stat, GSS_C_GSS_CODE);
     display_status_1(log, msg, min_stat, GSS_C_MECH_CODE);
}

/*
 * Create the kludge server response token
 *
 * Format: <username length>, <username>, <password length>, <password>
 *
 * where the lengths are two bytes each, most significant byte first.
 *
 * Note: if the username field is null, then the password represents
 * an error string.
 */
int create_kludge_token(username, password, token)
	char	*username;
	char	*password;
	gss_buffer_desc *token;
{
	int length, ulen, plen;
	char *buf, *cp;

	ulen = username ? strlen(username) : 0;
	plen = password ? strlen(password) : 0;
	length = ulen + plen + 4;
	buf = malloc(length);
	if (!buf)
		return ENOMEM;

	cp = buf;
	*cp++ = (char) ((ulen >> 8) & 0xFF);
	*cp++ = (char) (ulen & 0xFF);
	if (ulen) {
		strcpy(cp, username);
		cp += ulen;
	}

	*cp++ = (char) ((plen >> 8) & 0xFF);
	*cp++ = (char) (plen & 0xFF);
	if (plen)
		strcpy(cp, password);

	token->value = buf;
	token->length = length;
	
	return 0;
}

/*
 * Parse the kludge server response token
 */
int parse_kludge_token(username, password, token)
	char **username;
	char **password;
	gss_buffer_desc *token;
{
	char	*ubuf, *pbuf, *cp;
	int	len, ulen, plen;

	len = token->length;
	cp = token->value;
	ulen = (*cp++) << 8;
	ulen += *cp++;
	
	len = len - (ulen+1);
	if (len < 0)
		return -1;
	ubuf = malloc(ulen+1);
	if (!ubuf)
		return ENOMEM;
	strncpy(ubuf, cp, ulen);
	ubuf[ulen] = '\0';
	cp += ulen;

	plen = (*cp++) << 8;
	plen += *cp++;
	len = len - (plen+1);
	if (len < 0) {
		free(ubuf);
		return -1;
	}
	pbuf = malloc(plen+1);
	if (!pbuf) {
		free(ubuf);
		return ENOMEM;
	}
	strncpy(pbuf, cp, plen);
	pbuf[plen] = '\0';

	*username = ubuf;
	*password = pbuf;

	return 0;
}
