/*
 * gss-server.c --- originally taken from the Krb5 GSS Sample
 * application.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <ctype.h>

#include <gssapi/gssapi_generic.h>
#include "util.h"
#include "gss-misc.h"
#include "randpass.h"
#include "kludge.h"

#ifdef __alpha
#define KEYTAB_FILE "FILE:/etc/krb5.keytab.oracle"
#endif
#define LOG_FILE "/usr/tmp/kludge.log"
#define SECRET_FILE "/tmp/secret"

const char *log_file = NULL;
const char *srvtab_file = NULL;
const char *default_sid = NULL;

void usage()
{
     fprintf(stderr, "Usage: kludge-server [-port port] [-once] [-inetd]\n");
     fprintf(stderr, "       [-logfile file] [-srvtab file] [service_name]\n");
     exit(1);
}

void datestamp(FILE *log)
{	
	time_t	t;
	struct tm *tm;
	const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };


	t = time(0);
	tm = localtime(&t);

	fprintf(log, "%2d-%s-%4d %02d:%02d:%02d  ",
		tm->tm_mday, monstr[tm->tm_mon], tm->tm_year + 1900,
		tm->tm_hour, tm->tm_min, tm->tm_sec);
}

/*
 * Function: server_acquire_creds
 *
 * Purpose: imports a service name and acquires credentials for it
 *
 * Arguments:
 *
 * 	service_name	(r) the ASCII service name
 * 	server_creds	(w) the GSS-API service credentials
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 *
 * The service name is imported with gss_import_name, and service
 * credentials are acquired with gss_acquire_cred.  If either opertion
 * fails, an error message is displayed and -1 is returned; otherwise,
 * 0 is returned.
 */
static int server_acquire_creds(app_ctx, service_name, server_creds)
	kludge_t	app_ctx;
	char *service_name;
	gss_cred_id_t *server_creds;
{
	gss_buffer_desc name_buf;
	gss_name_t server_name;
	OM_uint32 maj_stat, min_stat;

	if (srvtab_file)
		setenv_if_not_set("KRB5_KTNAME", srvtab_file);

#ifdef KEYTAB_FILE
	setenv_if_not_set("KRB5_KTNAME", KEYTAB_FILE);
#endif

	name_buf.value = service_name;
	name_buf.length = strlen(name_buf.value) + 1;
	maj_stat = gss_import_name(&min_stat, &name_buf, 
				   (gss_OID) gss_nt_service_name,
				   &server_name);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(app_ctx->log, "importing name",
			       maj_stat, min_stat);
		return -1;
	}

	maj_stat = gss_acquire_cred(&min_stat, server_name, 0,
				    GSS_C_NULL_OID_SET, GSS_C_ACCEPT,
				    server_creds, NULL, NULL);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(app_ctx->log, "acquiring credentials",
			       maj_stat, min_stat);
		return -1;
	}

	(void) gss_release_name(&min_stat, &server_name);
	
	return 0;
}

/*
 * Function: server_establish_context
 *
 * Purpose: establishses a GSS-API context as a specified service with
 * an incoming client, and returns the context handle and associated
 * client name
 *
 * Arguments:
 *
 * 	s		(r) an established TCP connection to the client
 * 	service_creds	(r) server credentials, from gss_acquire_cred
 * 	context		(w) the established GSS-API context
 * 	client_name	(w) the client's ASCII name
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 *
 * Any valid client request is accepted.  If a context is established,
 * its handle is returned in context and the client name is returned
 * in client_name and 0 is returned.  If unsuccessful, an error
 * message is displayed and -1 is returned.
 */
int server_establish_context(app_ctx, context, client_name, ret_flags)
	kludge_t	app_ctx;
	gss_ctx_id_t *context;
	gss_buffer_t client_name;
	OM_uint32 *ret_flags;
{
	gss_buffer_desc send_tok, recv_tok;
	gss_name_t client;
	gss_OID doid;
	OM_uint32 maj_stat, min_stat;

	*context = GSS_C_NO_CONTEXT;
     
	do {
		if (recv_token(app_ctx->socket, &recv_tok) < 0) {
			fprintf(app_ctx->log, "failure reading token\n");
			return -1;
		}

		maj_stat =
			gss_accept_sec_context(&min_stat, context,
					       app_ctx->server_creds,
					       &recv_tok,
					       GSS_C_NO_CHANNEL_BINDINGS,
					       &client, &doid, &send_tok,
					       ret_flags, NULL, NULL);

		if (maj_stat != GSS_S_COMPLETE &&
		    maj_stat != GSS_S_CONTINUE_NEEDED) {
			display_status(app_ctx->log, "accepting context",
				       maj_stat, min_stat);
			(void) gss_release_buffer(&min_stat, &recv_tok);
			return -1;
		}

		(void) gss_release_buffer(&min_stat, &recv_tok);

		if (send_tok.length != 0) {
			if (send_token(app_ctx->socket, &send_tok) < 0) {
				fprintf(app_ctx->log,
					"failure sending token\n");
				return -1;
			}

			(void) gss_release_buffer(&min_stat, &send_tok);
		}
	} while (maj_stat == GSS_S_CONTINUE_NEEDED);

	maj_stat = gss_display_name(&min_stat, client, client_name, &doid);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(app_ctx->log, "displaying name",
			       maj_stat, min_stat);
		return -1;
	}
	maj_stat = gss_release_name(&min_stat, &client);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status(app_ctx->log, "releasing name",
			       maj_stat, min_stat);
		return -1;
	}
	return 0;
}

/*
 * Function: create_socket
 *
 * Purpose: Opens a listening TCP socket.
 *
 * Arguments:
 *
 * 	port		(r) the port number on which to listen
 *
 * Returns: the listening socket file descriptor, or -1 on failure
 *
 * Effects:
 *
 * A listening socket on the specified port and created and returned.
 * On error, an error message is displayed and -1 is returned.
 */
int create_socket(port)
     u_short port;
{
     struct sockaddr_in saddr;
     int s;
     int on = 1;
     
     saddr.sin_family = AF_INET;
     saddr.sin_port = htons(port);
     saddr.sin_addr.s_addr = INADDR_ANY;

     if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  perror("creating socket");
	  return -1;
     }
     /* Let the socket be reused right away */
     (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
     if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
	  perror("binding socket");
	  (void) close(s);
	  return -1;
     }
     if (listen(s, 5) < 0) {
	  perror("listening on socket");
	  (void) close(s);
	  return -1;
     }
     return s;
}

/*
 * Initialize the kludge server
 */
int init_kludge_server(app_ctx, service_name)
	char		*service_name;
	kludge_t	app_ctx;
{
	app_ctx->listen_socket = 0;
	app_ctx->socket = 0;

	if (!log_file)
		log_file = LOG_FILE;

	app_ctx->log = fopen(log_file, "a");
	if (!app_ctx->log) {
		perror(log_file);
		return -1;
	}

	randpass_init(SECRET_FILE);

	if (server_acquire_creds(app_ctx, service_name,
				 &app_ctx->server_creds) < 0)
		return -1;

	return 0;
}

/*
 * shutdown_kludge server
 */
void shutdown_kludge_server(app_ctx)
	kludge_t	app_ctx;
{
#ifdef DO_ORACLE
	char *error;
#endif
	OM_uint32 min_stat;

#ifdef DO_ORACLE
	Oracle_Shutdown(&error);
#endif
	if (app_ctx->log)
		fclose(app_ctx->log);		
	
	(void) gss_release_cred(&min_stat, &app_ctx->server_creds);
}

/*
 * Function: kludge_server
 *
 * Purpose: Performs the "kludge" service.
 *
 * Arguments:
 *
 * 	s		(r) a TCP socket on which a connection has been
 *			accept()ed
 * 	service_name	(r) the ASCII name of the GSS-API service to
 * 			establish a context as
 * 
 * Returns: -1 on error
 *
 * Effects:
 *
 * sign_server establishes a context, and performs a single sign request.
 *
 * A sign request is a single GSS-API sealed token.  The token is
 * unsealed and a signature block, produced with gss_sign, is returned
 * to the sender.  The context is the destroyed and the connection
 * closed.
 *
 * If any error occurs, -1 is returned.
 */
int kludge_server(app_ctx)
	kludge_t	app_ctx;
{
	gss_buffer_desc client_name, msg_buf, xmit_buf;
	gss_ctx_id_t context;
	OM_uint32 maj_stat, min_stat;
	int conf_state, ret_flags, retval;
	char	passbuf[256], *password=0, *username=0, *realm, *error=0;
	char 	*msg, tmpbuf[256], sid[32];
	int	len;
	
	xmit_buf.value = 0;
     
	/* Establish a context with the client */
	if (server_establish_context(app_ctx, &context,
				     &client_name, &ret_flags) < 0)
		return(-1);
	username = malloc(client_name.length+5);
	if (!username) {
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log, "No memory to allocate client name\n");
		return -1;
	}
	strcpy(username, "ops$");
	strncpy(username+4, client_name.value, client_name.length);
	username[client_name.length+4] = '\0';
	realm = strchr(username, '@');
	if (realm)
		*realm = 0;
	

	(void) gss_release_buffer(&min_stat, &client_name);

	retval = recv_sealed_token(app_ctx->socket, context, &msg_buf,
				   &conf_state, &maj_stat, &min_stat);
	if (retval) {
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log, "Couldn't get token: %s\n",
			strerror(retval));
		return(-1);
	}
	if (maj_stat != GSS_S_COMPLETE) {
	  	datestamp(app_ctx->log);
		display_status(app_ctx->log, "unsealing initial message",
			       maj_stat, min_stat);
		return(-1);
	}
	if (!conf_state) {
	  	datestamp(app_ctx->log);
		fprintf(stderr, "Warning!  Message not encrypted.\n");
	}

	/*
	 * check version code
	 */
	msg = msg_buf.value;
	if (msg[0] != 1) {
		sprintf(tmpbuf,
			"Protocol major version number %d not supported",
			msg[0]);
		error = string_duplicate(tmpbuf);
		if (error)
			goto fail_with_error;
		goto fail;
	}
	if (msg[1] > 1) {
		sprintf(tmpbuf,
			"Protocol minor version number %d not supported",
			msg[1]);
		error = string_duplicate(tmpbuf);
		if (error)
			goto fail_with_error;
		goto fail;
	}
	if (msg[1] == 0) {
		/*
		 * Old version, just set SID to be null, and use the
		 * default SID.
		 */
		sid[0] = 0;
		fprintf(app_ctx->log, "Connect protocol version 1.0\n");
	} else {
		len = msg[2];
		if (len > 16)
			len = 16;
		memcpy(sid, msg+3, len);
		sid[len] = 0;
		fprintf(app_ctx->log,
			"Connect protocol version 1.1+, sid='%s'\n", sid);
	}
	
	randpass_gen(passbuf, 8);
	password = passbuf;

	if (strchr(username, '/') ||
	    !realm || strcmp(realm+1, "ATHENA.MIT.EDU")) {
	  	datestamp(app_ctx->log);
		sprintf(tmpbuf, "Bad Kerberos principal: %s@%s\n", username,
			realm ? realm+1 : "???");
		error = string_duplicate(tmpbuf);
		if (error)
			goto fail_with_error;
		goto fail;
	}


#ifdef DO_ORACLE
	if (Oracle_Init(0, sid[0] ? sid : default_sid, &error)) {
		if (error)
			goto fail_with_error;
		else
			goto fail;
	}
	datestamp(app_ctx->log);
	fprintf(app_ctx->log, "Set password: '%s' for user '%s'\n", 
		password, username);
	Oracle_SetPassword(username, password, &error);
#endif

	if (error) {
	fail_with_error:
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log, "Failed to set password: %s\n", error);
		if (username) {
			free(username);
			username = 0;
		}
		password = error;
	}

	retval = create_kludge_token(username, password, &xmit_buf);
	if (retval) {
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log, "Error creating kludge token: %s\n",
			strerror(retval));
		return(-1);
	}
		
	conf_state = 1;
	retval = send_sealed_token(app_ctx->socket, context,
				   &xmit_buf, &conf_state,
				   &maj_stat, &min_stat);
	if (retval) {
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log, "Couldn't send token: %s\n",
			strerror(retval));
		return(-1);
	}
	if (maj_stat != GSS_S_COMPLETE) {
	  	datestamp(app_ctx->log);
		display_status(app_ctx->log, "sealing initial message",
			       maj_stat, min_stat);
		return(-1);
	}
	if (!conf_state) {
	  	datestamp(app_ctx->log);
		fprintf(app_ctx->log,
			"Warning!  Message not sent encrypted.\n");
	}

	retval = 0;

fail:
	/* Delete context */
	maj_stat = gss_delete_sec_context(&min_stat, &context, NULL);
	if (maj_stat != GSS_S_COMPLETE) {
	  	datestamp(app_ctx->log);
		display_status(app_ctx->log, "deleting context",
			       maj_stat, min_stat);
		return(-1);
	}

	if (username)
		free(username);
	if (error)
		free(error);
	free(xmit_buf.value);

	if (app_ctx->log)
		fflush(app_ctx->log);

	return(retval);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	char *service_name;
	u_short port = 17007;
	int s = 0;
	int once = 0;
	int do_inetd = 0;
	struct kludge appl_ctx;
	
	argc--; argv++;
	while (argc) {
		if (strcmp(*argv, "-port") == 0) {
			argc--; argv++;
			if (!argc) usage();
			port = atoi(*argv);
		} else if (strcmp(*argv, "-once") == 0) {
			once = 1;
		} else if (strcmp(*argv, "-inetd") == 0) {
			do_inetd = 1;
		} else if (strcmp(*argv, "-logfile") == 0) {
			argc--; argv++;
			if (!argc) usage();
			log_file = *argv;
		} else if (strcmp(*argv, "-srvtab") == 0) {
			argc--; argv++;
			if (!argc) usage();
			srvtab_file = *argv;
		} else if (strcmp(*argv, "-default-sid") == 0) {
			argc--; argv++;
			if (!argc) usage();
			default_sid = *argv;
		} else
			break;
		argc--; argv++;
	}
	if (argc != 1)
		usage();

	if ((*argv)[0] == '-')
		usage();
	
	service_name = *argv;

	if (init_kludge_server(&appl_ctx, service_name)  < 0)
		exit(1);

	if (do_inetd) {
		close(1);
		close(2);

		kludge_server(&appl_ctx);
		close(0);
	} else {
		int stmp;

		if ((stmp = create_socket(port))) {
			do {
				/* Accept a TCP connection */
				s = accept(stmp, NULL, 0);
				if (s < 0) {
					perror("accepting connection");
					continue;
				}
				appl_ctx.socket = s;
				/*
				 * this return value is not
				 * checked, because there's
				 * not really anything to do
				 * if it fails
				 */
				kludge_server(&appl_ctx);
			} while (!once);
		}
		close(stmp);
	}
	
	shutdown_kludge_server(&appl_ctx);

	/*NOTREACHED*/
	if (s)
		(void) close(s);
	return 0;
}
