/*
 * hmac_test.c --- hmac test routine
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "time.h"
#include "hmac_md5.h"

static void MDPrint(unsigned char [16]);

/*
 * The definition of an item.  If text is non-null, then text_len is
 * the length of text.  If text is NULL, then the item is text_len
 * characters of ch
 */
struct item {
	char *text;
	char ch;
	int  text_len;
};

struct test_suite {
	struct item key;
	struct item data;
	unsigned char result[16];
};

struct test_suite hmac_test_suite[] = {
	{ {  0, 0x0B, 16 },
	  {  "Hi There", 0, 8 },
	  0x92, 0x94, 0x72, 0x7a, 0x36, 0x38, 0xbb, 0x1c,
	  0x13, 0xf4, 0x8e, 0xf8, 0x15, 0x8b, 0xfc, 0x9d },

	{ { "Jefe", 0, 4 },
	  { "what do ya want for nothing?", 0, 28 },
	  0x75, 0x0c, 0x78, 0x3e, 0x6a, 0xb0, 0xb5, 0x03,
	  0xea, 0xa8, 0x6e, 0x31, 0x0a, 0x5d, 0xb7, 0x38 },

	{ { 0, 0xAA, 16 },
	  { 0, 0xDD, 50 },
	  0x56, 0xbe, 0x34, 0x52, 0x1d, 0x14, 0x4c, 0x88,
	  0xdb, 0xb8, 0xc7, 0x33, 0xf0, 0xe8, 0xb3, 0xf6 },

	{ { "\001\002\003\004\005\006\007\010\011\012\013\014\015\016\017\020"
	    "\021\022\023\024\025\026\027\030\031", 0, 25},
	  { 0, 0xCD, 50 },
          0x69, 0x7e, 0xaf, 0x0a, 0xca, 0x3a, 0x3a, 0xea,
	  0x3a, 0x75, 0x16, 0x47, 0x46, 0xff, 0xaa, 0x79 },

	{ { 0, 0x0C, 16 },
	  { "Test With Truncation", 0, 20 },
          0x56, 0x46, 0x1e, 0xf2, 0x34, 0x2e, 0xdc, 0x00,
	  0xf9, 0xba, 0xb9, 0x95, 0x69, 0x0e, 0xfd, 0x4c },

	{ { 0, 0xAA, 80 },
	  { "Test Using Larger Than Block-Size Key - Hash Key First", 0, 54 },
          0x6b, 0x1a, 0xb7, 0xfe, 0x4b, 0xd7, 0xbf, 0x8f,
	  0x0b, 0x62, 0xe6, 0xce, 0x61, 0xb9, 0xd0, 0xcd }, 

	{ { 0, 0xAA, 80 },
	  { "Test Using Larger Than Block-Size Key and Larger "
	    "Than One Block-Size Data", 0, 73 },
          0x6f, 0x63, 0x0f, 0xad, 0x67, 0xcd, 0xa0, 0xee,
	  0x1f, 0xb1, 0xf5, 0x62, 0xdb, 0x3a, 0xa5, 0x3e }, 

	{ { 0, 0, 0 },
	  { 0, 0, 0 }, }
};

void set_item(struct item *p, unsigned char *buf, int *len)
{
	if (p->text_len <= 0) {
		*len = 0;
		return;
	}
	if (p->text)
		memcpy(buf, p->text, p->text_len);
	else
		memset(buf, p->ch, p->text_len);
	*len = p->text_len;
}

void do_test(struct test_suite *t, int num)
{
	HMAC_MD5_CTX	context;
	unsigned char	keybuf[128], databuf[128], digest[16];
	int		keylen, datalen;

	memset(keybuf, 0, sizeof(keybuf));
	memset(databuf, 0, sizeof(databuf));
	set_item(&t->key, keybuf, &keylen);
	set_item(&t->data, databuf, &datalen);

	HMAC_MD5_SetKey(&context, keybuf, keylen);
	HMAC_MD5(&context, databuf, datalen, digest);

	printf("Test #%d digest is ", num);
	MDPrint(digest);
	printf("\n");

	if (memcmp(digest, t->result, 16)) {
		printf("Test #%d failed!  Should have been:  ", num);
		MDPrint(t->result);
		printf("\n\n");
	}
}

#define NUM_TRIALS 100000

static void HMAC_time_trial()
{
	HMAC_MD5_CTX	context;
	unsigned char	keybuf[128], databuf[1024], digest[16];
	time_t		startTime, endTime;
	int		i;

	memset(keybuf, 0xff, sizeof(keybuf));
	memset(databuf, 0x80, sizeof(databuf));

	printf("\n\nHMAC-MD5 time trial.  MAC'ing %d %d-byte blocks ...\n",
	       NUM_TRIALS, sizeof(databuf));

	/* Start timer */
	time (&startTime);

	HMAC_MD5_SetKey(&context, keybuf, sizeof(keybuf));

	for (i=0; i < NUM_TRIALS; i++) {
		HMAC_MD5(&context, databuf, sizeof(databuf), digest);
	}

	time (&endTime);

	printf ("\nTime = %ld seconds\n", (long)(endTime-startTime));
	printf("Speed = %ld bytes/second\n",
	       (long)sizeof(databuf) * (long)NUM_TRIALS/(endTime-startTime));
}

/* Prints a message digest in hexadecimal.
 */
static void MDPrint (digest)
unsigned char digest[16];
{
	unsigned int i;

	for (i = 0; i < 16; i++)
		printf ("%02x", digest[i]);
}

static void HMAC_test()
{
	int	i;

	for (i=0; i < 7; i++)
		do_test(&hmac_test_suite[i], i+1);
}

int main(argc, argv)
	int	argc;
	char	**argv;
{
	int	i;

	if (argc > 1) {
		for (i = 1; i < argc; i++)
			if (strcmp (argv[i], "-t") == 0)
				HMAC_time_trial();
			else if (strcmp (argv[i], "-x") == 0)
				HMAC_test();
	} else
		HMAC_test();
	exit(0);
}

